#!/usr/bin/env python3

from collections import OrderedDict
import datetime
import json
import matplotlib
import matplotlib.pyplot as plt
import multiprocessing
import numpy as np
import os
import os.path
from os.path import join
import subprocess
import types
from types import SimpleNamespace


def setup_outdir(name, base=None):
    """
    Setup and output directory.

    args:
        name        name of the output directory
        base        base directory (or file in base directory), it not passed,
                    the parent directory of the directory in which this file is
                    in is used (assuming that we are in a utils module which is
                    a directory lying in the project base folder)

    Best practice usage:

        setup_outdir("out", __file__)

    to get subfolder in directory of main script or

        setup_outdir("out", os.path.curdir)

    to get subfolder of current working directory.
    """

    if base is not None:
        if os.path.isfile(base):
            base = os.path.dirname(base)
        elif os.path.isdir(base):
            pass
        else:
            raise IOError("need to pass a base which is either a file or a" + \
                    "directory.")
    else:
        base = os.path.dirname(os.path.dirname(__file__))

    outdir = join(base, name)

    if not os.path.exists(outdir):
        os.makedirs(outdir)

    return outdir


def setup_numbered_outdir(basename, base=None, title=None):
    """
    Setup and a numbered output directory for an experiment.

    args:
        name        name of the output directory in which numbered directory is
                    created
        base        base directory (or file in base directory), it not passed,
                    the parent directory of the directory in which this file is
                    in is used (assuming that we are in a utils module which is
                    a directory lying in the project base folder)
        title       additional string to append to directory name

    Best practice usage: see documentation of setup_outdir().
    """
    outdir = setup_outdir(basename, base)

    dirs = [d for d in os.listdir(outdir) if os.path.isdir(join(outdir, d))]
    dirs_split = [d.split('_')[0] for d in dirs]

    dir_nums = []

    for d in dirs_split:
        try:
            num = int(d)
            dir_nums += [num]
        except ValueError:
            pass

    n = max(dir_nums) + 1 if len(dir_nums) > 0 else 0

    now = '{:%Y-%m-%d_%H-%M-%S}'.format(datetime.datetime.now())
    dirname = '{0:04d}_{1:s}{2:s}'.format(n, now, '_'+title if title else '')
    outdir = setup_outdir(join(outdir, dirname), base)

    return outdir


def setup_numpy_and_matplotlib(*, np_prec=3, plt_size=16, plt_family='sans', plt_usetex=False, seed=None, verbose=True):
    """
    Setup numpy and matplotlib:
        - numpy: set array printing precision
        - numpy: set print width according to current terminal size
        - numpy: suppress printing of small values in scientific notation
        - numpy: set seed
        - matplotlib: use tex
        - matplotlib: set font size
        - matplotlib: set 300 dpi for savefig

    keyword-only args:
        np_prec     numpy array printing precision (default: 3)
        plt_size    matplotlib font size (defaut: 16)
        seed        numpy seed (default: None, sets random seed)
        verbose     print additional info, default: True
    """

    np.set_printoptions(precision=np_prec, suppress=True)

    try:
        ret = subprocess.run(['stty', 'size'], stdout=subprocess.PIPE).stdout
        try:
            cols = int(ret.split()[1])

            np.set_printoptions(linewidth=cols)
        except IndexError:
            pass

    except FileNotFoundError:
        pass

    if seed is None:
        np.random.seed(seed)
        seed = np.random.randint(np.iinfo(np.int32).max)

    np.random.seed(seed)

    if verbose:
        print('numpy seeded with', seed)

    matplotlib.rcParams['savefig.dpi'] = 300.
    plt.rc('text', usetex=plt_usetex)
    plt.rc('font', family=plt_family, size=plt_size)


def get_vars():
    """
    Get a dict of global variables without functions, modules, or names
    starting with an underscore.
    """

    g = globals().copy()
    d = {}

    for k, v in g.items():
        if k.startswith("_") or \
            type(v) is types.FunctionType or \
            type(v) is types.ModuleType:
            continue

        d[k] = v

    return d


def print_git_status():
    print('git status:')

    rev = subprocess.run(['git', 'rev-parse', 'HEAD'], stdout=subprocess.PIPE).stdout
    stat = subprocess.run(['git', 'status'], stdout=subprocess.PIPE).stdout
    diff = subprocess.run(['git', 'diff'], stdout=subprocess.PIPE).stdout

    print('revision:', rev.decode('UTF-8'))
    print(stat.decode('UTF-8'))
    print(diff.decode('UTF-8'))


def shell_exec(command, check=False):
    run = subprocess.run(command, check=check, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    return run.stdout.decode('utf-8'), run.stderr.decode('utf-8'),


def shell_exec_with_pipe(command_list):
    assert type(command_list) is list
    assert len(command_list) > 0

    ps_ = []

    for cmd in command_list[:-1]:
        stdin = None if len(ps_) == 0 else ps_[-1].stdout
        ps = subprocess.Popen(cmd, stdout=subprocess.PIPE, stdin=stdin)
        ps_ += [ps]

    output = subprocess.check_output(command_list[-1], stdin=ps.stdout)

    [ps.wait() for ps in ps_]

    return output.decode('utf-8').strip()


def shell_exec_silent(command):
    return subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE).returncode


def dict_to_string(d):
    assert type(d) in [dict, OrderedDict]

    if type(d) is OrderedDict:
        s = '_'.join(str(k)+str(v)
                for k, v in d.items())
    else:
        s = '_'.join(str(k)+str(d[k])
                for k in sorted(d.keys()))

    return s


def dump_dict(data, *, dumpfile=None, append=False, key='', ordered=True, indent=4, print_stdout=True):
    """
    Write a dict to stdout and/or a file.

    args:
        data            dict to write

    keyword-only args:
        dumpfile        if set, write to this file
        append          append to dumpfile, otherwise overwrite (default:
                        False)
        key             string to print as first line (default: "", no
                            printing is performed)
        ordered         order dict keys (default: True)
        indent          number of spaces for indentation (default: 4)
        print_stdout    print to stdout (default: True)
    """

    if type(data) is SimpleNamespace:
        data = data.__dict__

    s = json.dumps(data, indent=indent, sort_keys=ordered)

    s_file = s
    s_stdout = s

    if len(key) > 0:
        s_file = "// " + key + "\n" + s
        s_stdout = key + ":\n" + s

    mode = 'a' if append else 'w'

    if dumpfile is not None:
        with open(dumpfile, mode) as f:
            f.write(s_file)
            f.write('\n')

    if print_stdout:
        print(s_stdout)


def join_nested_lists(lists, level, *, flatten_inner=True):
    """
    Join nested lists at a certain level.

    args:
        lists           lists to join
        level           nestation level at which lists are to join

    keyword-only args:
        flatten_inner   flatten at join level instead of introducing
                        a new nesting level, default: True

    returns
        joined_list
    """

    assert level >= 0

    def __recur(lists, level, flatten_inner):
        if level == 0:
            if flatten_inner:
                return [elem for l in lists for elem in l]
            else:
                return list(lists)
        else:
            for l in lists[1:]:
                assert len(l) == len(lists[0]), 'inner dimension mismatch'
            return [__recur(l_, level-1, flatten_inner) for l_ in zip(*lists)]

    return __recur(lists, level, flatten_inner)


def print_stats(a, title=None, fmt='{:.2g}', median=True, arange=True, percentiles=False, binary=False):
    title = '' if title is None else title
    a = np.asarray(a)

    if a.size == 0:
        print('{:s}: empty'.format(title))
        return

    if len(a.ravel()) == 1:
        print(('{:s}: '+fmt+' (scalar)').format(title, a.ravel()[0]))
        return

    stats = []

    if not binary:
        s = '{:s}: ' + fmt + ' +/- ' + fmt
        stats += [a.mean(), a.std()]

        in_paren = False

        if median:
            s += ' (median: ' + fmt
            in_paren = True
            stats += [np.median(a)]

        if arange:
            if not in_paren:
                s += ' ('
                in_paren = True
            else:
                s += ', '
            s += 'range: ' + fmt + ' -- ' + fmt
            stats += [a.min(), a.max()]

        if percentiles:
            if not in_paren:
                s += ' ('
                in_paren = True
            else:
                s += ', '
            s += '5--95 percentile: ' + fmt + ' -- ' + fmt
            stats += [*np.percentile(a, [5, 95])]

        if in_paren:
            s += ')'
    else:
        s = '{:s}: {:d}/{:d}'
        if fmt != '':
            s += ' (' + fmt + ')'
        stats = [a.sum(), len(a.reshape(-1)), a.mean()]

    s = s.format(title, *stats)
    print(s)

    return stats


def get_hyphenated_ranges(x_):
    l = [str(x) for x in sorted(x_)]

    ranges = []

    for x in l:
        if not ranges:
            ranges.append([x])
        elif int(x)-prev_x == 1:
            ranges[-1].append(x)
        else:
            ranges.append([x])
        prev_x = int(x)

    ranges_joined = ['-'.join([r[0], r[-1]] if len(r) > 1 else r) for r in ranges]

    return ','.join(ranges_joined)


def multi_logical_and(masks):
    assert isinstance(masks, (tuple, list))
    assert len(masks) >= 2
    for mask in masks: assert isinstance(mask, np.ndarray), 'all masks must be numpy arrays'
    shape = masks[0].shape
    for mask in masks[1:]: assert shape == mask.shape, 'all masks must have the same shape'

    base = np.ones_like(masks[0])
    for mask in masks:
        base = (base & mask)

    return base


def multi_logical_or(masks):
    assert isinstance(masks, (tuple, list))
    assert len(masks) >= 2
    for mask in masks: assert isinstance(mask, np.ndarray), 'all masks must be numpy arrays'
    shape = masks[0].shape
    for mask in masks[1:]: assert shape == mask.shape, 'all masks must have the same shape'

    base = np.zeros_like(masks[0])
    for mask in masks:
        base = (base | mask)

    return base


def format_time(seconds):
    if seconds < 1e-2:
        s = '{0:g}'.format(seconds)+' s'
    elif seconds < 10:
        s = '{0:.2f}'.format(seconds)+' s'
    elif seconds < 60:
        s = '{0:.1f}'.format(seconds)+' s'
    elif seconds < 3600:
        s = '{0:.2f}'.format(seconds/60)+' min'
    elif seconds < 86400:
        s = '{0:.2f}'.format(seconds/3600)+' h'
    else:
        s = '{0:.2f}'.format(seconds/86400)+' days'

    return s


def projected_time_left(iters_total, iters_done, iter_durations, use_last=100, median=False):
    if median:
        iter_time = np.median(iter_durations[-use_last:])
    else:
        iter_time = np.mean(iter_durations[-use_last:])
    return format_time((iters_total - iters_done) * iter_time)
