#!/usr/bin/env python3

from matplotlib.patches import Rectangle
import matplotlib.pyplot as plt
import multiprocessing as mp
import numpy as np
import sys
import warnings


# import nest without verbose text
if 'nest' not in sys.modules:
    print('importing nest with --quiet')
    sys.argv.append('--quiet')

exec_name = sys.argv[0]
import nest
sys.argv[0] = exec_name


def setup_nest(seeds=None, modname=None, dt=.1, print_time=False, thread_offs=0, thread_div=1, verbose=True):
    """
    Setup nest: define number of cpus to use, set random seeds, load modules.

    args:
        print_time      enable nest time printing (default: False)
        thread_offs     number of cores not to use, default: 0
        thread_div      number of partitions for rest of cores, i.e. number of
                        threads is #CPUs / thread_div, if 0, do not use
                        multithreading, default: 1
        modname         if set, load this module, can also be a list
        verbose         print info on settings and actions (default: True)
    """

    if not verbose:
        nest.set_verbosity('M_DEPRECATED')

    nest.ResetKernel()

    # nest kernel settings
    if thread_div > 0:
        num_cpus = mp.cpu_count() - thread_offs
        num_thread = int(num_cpus / thread_div)
    else:
        num_thread = 1
    cfg = {'print_time': print_time, 'local_num_threads': num_thread, 'resolution': dt}
    nest.SetKernelStatus(cfg)

    # set random seeds
    if seeds is None:
        num_seeds = len(nest.GetKernelStatus()['rng_seeds']) + 1
        seeds = np.random.choice(1000, size=num_seeds, replace=False).tolist()
    elif type(seeds) is int:
        num_seeds = len(nest.GetKernelStatus()['rng_seeds']) + 1
        seeds = np.arange(seeds, num_seeds+seeds).tolist()
    else:
        assert len(seeds) == len(nest.GetKernelStatus()['rng_seeds']) + 1

    nest.SetKernelStatus({'grng_seed': seeds[0]})
    nest.SetKernelStatus({'rng_seeds': seeds[1:]})

    # load modules
    if modname is not None:
        if type(modname) is str:
            modname = [modname]

        for name in modname:
            load_module(name)

    if verbose:
        print('nest using {0:d} threads'.format(num_thread))

        if modname is not None:
            print('loaded nest modules:' + " ".join(modname))

        print('nest seeded with', seeds)


def load_module(modname):
    err = "DynamicModuleManagementError in Install: Module '{0:s}' is " + \
            "loaded already."

    try:
        nest.Install(modname)
    except nest.NESTError as ex:
        err_modname = err.format(modname)

        if ex.args[0] == err_modname:
            pass  # module already loaded
        else:
            raise ex

def get_spikes_from_spike_recorder(spike_recorder, population=None, t_lim=None, count=None, random=False):
    """
    Extract spikes from spike recorder and perform some optional processing.

    args:
        spike_recorder      NEST spike recorder
        population          targets of spike recorder, can be None if count
                            parameter is not used
        t_lim               optional, tuple of start and end time for
                            extracting only spikes within a certain time
                            range (useful for plotting)
        count               optional, number of neurons to extract spikes
                            from
        random              optional, only used if count is not None, bool
                            indicating whether the number of neurons to be
                            extract should be random, default: False

    returns:
        times               extracted spike times
        senders             extracted spike sender ids
    """
    events = nest.GetStatus(spike_recorder)[0]['events']
    times = events['times']
    senders = events['senders']

    if count is not None:
        ids = np.arange(min(population), max(population)+1)

        if not random:
            id_selection = ids[:count]
        else:
            id_selection = np.random.choice(ids, count, replace=False)

        mask = np.isin(senders, id_selection)

        times = times[mask]
        senders = senders[mask]

    if t_lim is not None:
        times, senders = filter_spikes(times, senders, t_lim=t_lim)

    return times, senders

def get_spike_times_by_id(input, *, pop=None, unit_s=False):
    """
    Separate the spike times per neuron.

    args:
        input               either nest device for input or list of length 2
                            containing times and senders

    keyword-only args:
        pop                 nest neuron ids, if None (default), the output will
                            only contain neurons which have fired
        unit_s              return spike times in s, unlike the NEST default
                            ms, default: False

    returns:
        pop                 vector containing nest neuron ids
        spikes_per_neuron   list containing a list of spike times for each
                            neuron in ms
    """

    if len(input) == 1:
        times, senders = get_spikes_from_spike_recorder(input)
    elif len(input) == 2:
        times, senders = input
    else:
        raise ValueError('bad input passed, length is neither 1 or 2.')

    if len(senders) == 0:
        return [], []

    if pop is None:
        min_id, max_id = min(senders), max(senders)
        pop = np.arange(min_id, max_id+1)

    spikes_per_id = []
    scale = 1000. if unit_s else 1.

    for id in pop:
        m = (senders == id)
        spikes_per_id += [times[m] / scale]

    return pop, spikes_per_id

def get_weights_from_conn(connections):
    """
    Retrieve weights from a list of Nest connections.

    args:
        connections     list obtained using Nest.GetConnection()

    returns
        weights         numpy.ndarray with weights
    """
    weights = [s['weight'] for s in nest.GetStatus(connections)]

    return np.asarray(weights)

def get_weights_from_population(population_in, population_out):
    """
    Extract weights between two populations of neurons.

    args:
        population_in       presynaptic neurons
        population_out      postsynaptic neurons

    returns:
        weights             matrix of shape (N, 3) if there are N connections
                            in total between the two populations; the first
                            column contains the presynaptic neuron id, the
                            second column contains the postsynaptic neuron id,
                            and the third column contains the weight
    """
    conn = nest.GetConnections(population_in, population_out)
    stats = nest.GetStatus(conn)

    weights = np.zeros((len(stats), 3))

    for k in range(len(stats)):
        source = conn[k][0]
        target = conn[k][1]
        weight = stats[k]['weight']

        weights[k,:] = [source, target, weight]

    return weights

def get_wiring(pop_in, pop_out, connections=None):
    wiring = []

    if connections is None:
        connections = nest.GetConnections(pop_in, pop_out)

    status = nest.GetStatus(connections)

    for c, s in zip(connections, status):
        index_in = pop_in.index(c[0])
        index_out = pop_out.index(c[1])

        w = s['weight']
        wiring += [(index_in, index_out, w)]

    return wiring

def restore_wiring(pop_in, pop_out, wiring, syn_spec, scale=1.):
    # create target unit vectors
    target_in = [pop_in[w[0]] for w in wiring]
    target_out = [pop_out[w[1]] for w in wiring]
    weights = [scale*w[2] for w in wiring]

    syn_spec['weight'] = weights

    # connect one-to-one
    nest.Connect(target_in, target_out, {'rule': 'one_to_one'}, syn_spec)

def filter_spikes(times, senders, *, targets=None, t_lim=None):
    """
    Filter spikes extracted from a spike recorder.

    args:
        times               spike times
        senders             spike sender ids

    keyword-only args:
        targets             if set, extract only neurons which have an id given
                            in this list
        t_lim               optional, tuple of start and end time for
                            extracting only spikes within a certain time range
                            (useful for plotting)

    returns:
        times               filtered spike times
        senders             filtered spike sender ids

    """
    if targets is not None:
        mask = np.isin(senders, targets)

        times = times[mask]
        senders = senders[mask]

    if t_lim is not None:
        assert(len(t_lim) == 2)

        mask = (times > t_lim[0]) & (times <= t_lim[1])

        times = times[mask]
        senders = senders[mask]

    return [times, senders]

def reorder_neurons(times, senders, targets):
    """
    Reorder neurons extracted from spike recorder according to given list of
    target ids.

    args:
        times               spike times
        senders             spike sender ids
        targets             list of ids, should contain the same elements as
                            the current neuron ids given in senders; the ids
                            will re-assigned in ascending order according to
                            the order in this list

    returns:
        times               spike times
        senders             re-arranged senders
    """
    assert(min(senders) > len(targets))

    times, senders = filter_spikes(times, senders, targets=targets)

    d = {}
    for k, t in enumerate(targets):
        d[t] = k

    return times, np.vectorize(d.__getitem__)(senders)

def neuron_spike_rates(input, t_lim, pop):
    """
    Compute the spike rate per neuron from spike detector data.

    args:
        input               either nest device for input or list of length 2
                            containing times and senders
        t_lim               tuple of start and end time in ms for rate
                            calculation
        pop                 nest neuron ids

    returns:
        rates               firing rates for returned population in Hz
    """

    assert len(t_lim) == 2
    dt = t_lim[1] - t_lim[0]

    if len(input) == 1:
        times, ids = get_spikes_from_spike_recorder(input)
    elif len(input) == 2:
        times, ids = input
    else:
        raise ValueError('bad input passed, length is neither 1 or 2.')

    # filter time as passed
    times, ids = filter_spikes(times, ids, t_lim=t_lim)

    # count spikes in data
    spike_ids, spike_counts = np.unique(ids, return_counts=True)

    # filter to spikes in pop
    mask = np.isin(spike_ids, pop)
    spike_ids, spike_counts = spike_ids[mask], spike_counts[mask]

    # save counts
    N = max(pop) + 1
    counts = np.zeros(N)
    counts[spike_ids] = spike_counts

    # get rates
    rates = counts / dt * 1000.

    # filter to match pop
    mask = np.isin(np.arange(N), pop)
    rates = rates[mask]

    return rates

def spike_hist(input, num_neurons, *, t_lim=None, binsize=10.):
    """
    Compute the mean firing rate over some time window.

    args:
        input               spike detector (if input is of type tuple or list
                            and has length 1) or [times, senders] (if input is
                            of type tuple or list and has length 2), otherwise,
                            an exception is raised
        num_neurons         number of neurons in input

    keyword-only args:
        t_lim               tuple of start and end time, if None (default),
                            time of first and last spike is used
        binsize             bin size for rate computation

    returns:
        times               vector of bin center times
        rates               vector of firing rates
    """

    times, senders = __get_input(input)

    dt = binsize / 2.

    if t_lim is None:
        t0, t1 = min(times), max(times)
    else:
        t0, t1 = t_lim

    t_ = np.arange(t0, t1+dt, dt)
    t_bin0_, t_bin1_ = t_-dt, t_+dt

    spike_count_ = []

    for t_bin0, t_bin1 in zip(t_bin0_, t_bin1_):
        spike_count_ += [len(filter_spikes(times, senders, t_lim=(t_bin0, t_bin1))[0])]

    rate_ = np.asarray(spike_count_) / num_neurons / binsize * 1000.  # in Hz

    return t_, rate_


def simulate_and_execute(t_sim, interval, func, const_output_length=False, eval_at_start=True):
    """
    Run a NEST simulation and periodically execute a function. The first call
    to func is performed before simulating.

    args:
        t_sim               total simulation time
        interval            time per iteration after which func is executed
        func                function to call every interval, must return an
                            iterable
        const_output_length if True, try to pre-allocate a numpy array for the
                            output data, in this case func must return a numpy
                            array whose size and data type will be used for the
                            output array, default: False
        eval_at_start       evaluate func before simulating, default: True

    returns:
        t                   vector NEST simulation times when func was executed
        data                return values of func at given times
    """
    t_start = nest.GetKernelStatus('time')
    t_end = t_start + t_sim
    N = int((t_end - t_start) / interval) + int(eval_at_start)

    t_ = np.zeros(N)
    t_cur = t_start

    if not eval_at_start:
        nest.Simulate(interval)

    data = func()
    n = 1
    t_[n-1] = t_cur

    if not const_output_length:
        data_ = [data]
    else:
        assert type(data) is np.ndarray

        data_ = np.zeros((N, *data.shape), dtype=data.dtype)
        data_[n-1,:] = data

    while t_cur < t_end:
        nest.Simulate(interval)

        data = func()
        n += 1

        t_cur = nest.GetKernelStatus('time')
        t_[n-1] = t_cur

        if not const_output_length:
            data_ += [data]
        else:
            data_[n-1,:] = data

    if n != N:
        warnings.warn("not all value in the data array in simulate_and_execute()"+
                " were filled", RuntimeWarning)

    return t_, data_

def plot_spikes(input, *, targets=None, title='', size=5., xlabel='time', ylabel='neuron', xlim=None, ylim=None, markers=[], marker_color_div=1, marker_style='--', marker_width=1., xticks=None, yticks=False, subplot=False, sharex=None, wide=False, height_scale=1., hide_spines_left=False, hide_spines_right=False, highlight=[], save=None, close=False):
    """
    args:
        input               spike detector (if input is of type tuple or list
                            and has length 1) or [times, senders] (if input is
                            of type tuple or list and has length 2), otherwise,
                            an exception is raised

    keyword-only args:
        targets             None or a list of neurons to filter and solely
                            display
        title               title of plot
        size                size of spike marker, default: 20
        xlabel              label for x-axis
        ylabel              label for y-axis
        xlim                None or tuple (start_time, end_time) of plot
        ylim                None or tuple (min_neuron_id, max_neuron_id)
        markers             list of times where colored vertical lines will be
                            placed
        marker_color_div    number of consecutive markers sharing the same
                            color, default: 1
        marker_style        pyplot line style for markers, default: '--'
        marker_width        pyplot line width for markers, default: 1.
        xticks              None or list of tuples (time, string), then,
                            instead of normal xticks, the strings are placed
                            as ticks at the given times
        yticks              bool, whether to show yticks for neuron ids, or
                            list of tuples with [value, label], defaut: False
        subplot             if False (default), create a new figure, if True,
                            do not, if value is a 3-tuple, create a subplot with
                            given values, in the latter two cases, the values
                            for wide and height_scale are ignored wide
                            make double width plot, default: False
        sharex              x-axis to share, default: None
        wide                if True, plot in double width
        height_scale        scale height, default: 1
        hide_spines_left    hide plot spine on left for a broken-axis plot
        hide_spines_right   hide plot spine on right for a broken-axis plot
        highlight           list of 5-tuples, each defining a field to
                            highlight, [min_neuron_id, max_neuron_id, x_start,
                            x_end, color]
        save                None or file name for saving
        close               whether to close the figure after drawing it
    """

    # TODO: use __get_input, __setup_subplot, and __post_plot

    if type(input) in [tuple, list] and len(input) == 1:
        ev = nest.GetStatus(input)[0]['events']
        times = ev['times']
        ids = ev['senders']

    elif type(input) in [tuple, list] and len(input) == 2:
        times, ids = input

    elif type(input) in [tuple, list]:
        err = 'plot_spikes: input has wrong length ({0:d}).'
        raise ValueError(err.format(len(input)))

    else:
        err = 'plot_spikes: input has wrong data type ({0:s}).'
        raise ValueError(err.format(str(type(input))))

    if targets is not None:
        mask = np.in1d(ids, targets)

        times = times[mask]
        ids = ids[mask]

        # change ids
        ids_old = ids.copy()
        for k, target in enumerate(targets):
            ids[ids_old == target] = k

    if subplot is False:
        width = 12. if wide else 6.
        height = 4. * height_scale
        plt.figure(figsize=(width, height))
        plt.subplot(1, 1, 1, sharex=sharex)

    elif type(subplot) in [tuple, list] and len(subplot) == 3:
        plt.subplot(*subplot, sharex=sharex)

    elif subplot is True:
        pass

    else:
        raise ValueError('bad value for subplot passed:', subplot)

    plt.scatter(times, ids, s=size, marker=2, color='#777777')

    for k, m in enumerate(markers):
        color = 'C{0:d}'.format(k // marker_color_div)
        plt.axvline(x=m, c=color, lw=marker_width, linestyle=marker_style)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)

    if xticks is not None:
        plt.gca().set_xticks([x[0] for x in xticks])
        plt.gca().set_xticklabels([x[1] for x in xticks])
    else:
        plt.locator_params(axis='x', nbins=3)

    if type(yticks) is bool:
        if not yticks:
            plt.yticks([])
        else:
            plt.locator_params(axis='y', nbins=3)
    else:
        plt.gca().set_yticks([x[0] for x in yticks])
        plt.gca().set_yticklabels([x[1] for x in yticks])

    if xlim is not None:
        assert(len(xlim) == 2)
        plt.xlim(*xlim)

    if ylim is not None:
        assert(len(ylim) == 2)
        plt.ylim(*ylim)

    if hide_spines_left:
        plt.gca().spines['left'].set_visible(False)

    if hide_spines_right:
        plt.gca().spines['right'].set_visible(False)

    for hl in highlight:
        start = [hl[2], hl[0]]
        width = hl[3]-hl[2]
        height = hl[1]-hl[0]
        rect = Rectangle(start, width, height, zorder=-10, alpha=0.2, facecolor=hl[4])
        plt.gca().add_patch(rect)

    plt.tight_layout()

    if save is not None:
        plt.savefig(save)

    if close:
        plt.close()

def plot_spike_hist(input, num_neurons, *, binsize=10., ylabel='rate', yticks=True, **kwargs):
    """
    args:
        input               spike detector (if input is of type tuple or list
                            and has length 1) or [times, senders] (if input is
                            of type tuple or list and has length 2), otherwise,
                            an exception is raised
        num_neurons         number of neurons in input

    keyword-only args:
        binsize             bin size for rate computation
        ylabel              label for y-axis, default: 'rate'
        yticks              bool, whether to show yticks for neuron ids, or
                            list of tuples with [value, label], defaut: True

    additional args: see docstrings of __setup_subplot and __post_plot
    functions
    """

    if 'xlim' in kwargs.keys():
        t_lim = kwargs['xlim']
    else:
        t_lim = None

    t_, rate_ = spike_hist(input, num_neurons, t_lim=t_lim, binsize=binsize)

    # plot

    __setup_subplot(**kwargs)

    plt.plot(t_, rate_)

    kwargs['ylabel'] = ylabel
    kwargs['yticks'] = yticks

    __post_plot(**kwargs)

def __get_input(input):
    """
    Retrieve input from data lists or spike detector.

    args:
        input               spike detector (if input is of type tuple or list
                            and has length 1) or [times, senders] (if input is
                            of type tuple or list and has length 2), otherwise,
                            an exception is raised

    returns:
        times
        senders
    """

    if type(input) in [tuple, list] and len(input) == 1:
        ev = nest.GetStatus(input, 'events')[0]
        times = ev['times']
        senders = ev['senders']

    elif type(input) in [tuple, list] and len(input) == 2:
        times, senders = input

    elif type(input) in [tuple, list]:
        err = 'plot_spikes: input has wrong length ({0:d}).'
        raise ValueError(err.format(len(input)))

    else:
        err = 'plot_spikes: input has wrong data type ({0:s}).'
        raise ValueError(err.format(str(type(input))))

    return times, senders

def __setup_subplot(subplot=False, sharex=None, wide=False, height_scale=1., **kwargs):
    """
    Setup subplot for other plot functions.

    args:
        subplot             if False (default), create a new figure, if True,
                            do not, if value is a 3-tuple, create a subplot with
                            given values, in the latter two cases, the values
                            for wide and height_scale are ignored wide
                            make double width plot, default: False
        sharex              x-axis to share, default: None
        wide                if True, plot in double width
        height_scale        scale height, default: 1
    """

    if subplot is False:
        width = 12. if wide else 6.
        height = 4. * height_scale
        plt.figure(figsize=(width, height))
        plt.subplot(1, 1, 1, sharex=sharex)

    elif type(subplot) in [tuple, list] and len(subplot) == 3:
        plt.subplot(*subplot, sharex=sharex)

    elif subplot is True:
        pass

    else:
        raise ValueError('bad value for subplot passed:', subplot)

def __post_plot(title='', xlabel='time', ylabel='neuron', xlim=None, ylim=None, markers=[], marker_color_div=1, marker_style='--', marker_width=1., xticks=None, yticks=False, hide_spines_left=False, hide_spines_right=False, highlight=[], save=None, close=False, **kwargs):
    """
    Style plot after main plot functions have been called.

    args:
        title               title of plot
        xlabel              label for x-axis
        ylabel              label for y-axis
        xlim                None or tuple (start_time, end_time) of plot
        ylim                None or tuple (min_neuron_id, max_neuron_id)
        markers             list of times where colored vertical lines will be
                            placed
        marker_color_div    number of consecutive markers sharing the same
                            color, default: 1
        marker_style        pyplot line style for markers, default: '--'
        marker_width        pyplot line width for markers, default: 1.
        xticks              None or list of tuples (time, string), then,
                            instead of normal xticks, the strings are placed
                            as ticks at the given times
        yticks              bool, whether to show yticks for neuron ids, or
                            list of tuples with [value, label], defaut: False
        hide_spines_left    hide plot spine on left for a broken-axis plot
        hide_spines_right   hide plot spine on right for a broken-axis plot
        highlight           list of 5-tuples, each defining a field to
                            highlight, [min_neuron_id, max_neuron_id, x_start,
                            x_end, color]
        save                None or file name for saving
        close               whether to close the figure after drawing it
    """

    for k, m in enumerate(markers):
        color = 'C{0:d}'.format(k // marker_color_div)
        plt.axvline(x=m, c=color, lw=marker_width, linestyle=marker_style)

    plt.xlabel(xlabel)
    plt.ylabel(ylabel)
    plt.title(title)

    if xticks is not None:
        plt.gca().set_xticks([x[0] for x in xticks])
        plt.gca().set_xticklabels([x[1] for x in xticks])
    else:
        plt.locator_params(axis='x', nbins=3)

    if type(yticks) is bool:
        if not yticks:
            plt.yticks([])
        else:
            plt.locator_params(axis='y', nbins=3)
    else:
        plt.gca().set_yticks([x[0] for x in yticks])
        plt.gca().set_yticklabels([x[1] for x in yticks])

    if xlim is not None:
        assert(len(xlim) == 2)
        plt.xlim(*xlim)

    if ylim is not None:
        assert(len(ylim) == 2)
        plt.ylim(*ylim)

    if hide_spines_left:
        plt.gca().spines['left'].set_visible(False)

    if hide_spines_right:
        plt.gca().spines['right'].set_visible(False)

    for hl in highlight:
        start = [hl[2], hl[0]]
        width = hl[3]-hl[2]
        height = hl[1]-hl[0]
        rect = Rectangle(start, width, height, zorder=-10, alpha=0.2, facecolor=hl[4])
        plt.gca().add_patch(rect)

    plt.tight_layout()

    if save is not None:
        plt.savefig(save)

    if close:
        plt.close()

class InputUnit:
    def __init__(self, targets, rule_dict=None, syn_spec=None, num_units=1, rate=75):
        # save args
        self._targets = targets
        self._rule_dict = rule_dict
        self._syn_spec = syn_spec
        self._num_units = num_units

        # setup
        self._generator = nest.Create('poisson_generator')
        self._parrots = nest.Create('parrot_neuron', num_units)
        nest.Connect(self._generator, self._parrots)
        nest.Connect(self._parrots, targets, rule_dict, syn_spec)

        self._spike_detector = nest.Create('spike_detector')
        nest.Connect(self._parrots, self._spike_detector)

        # defaults
        self._on = None
        self._start = nest.GetStatus(self._generator, 'start')
        self._stop = nest.GetStatus(self._generator, 'stop')

        # set using setter
        self.rate = rate

    def get_spikes(self, normalize_ids=False):
        ev = nest.GetStatus(self._spike_detector, 'events')[0]
        times, ids = ev['times'], ev['senders']

        if normalize_ids:
            ids -= min(self._parrots)

        return times, ids

    @property
    def rate(self):
        return self._rate

    @rate.setter
    def rate(self, value):
        value = float(value)
        nest.SetStatus(self._generator, {'rate': value})
        self._rate = value

    @property
    def start(self):
        return self._start

    @rate.setter
    def start(self, value):
        if self._on is not None:
            warnings.warn('start: overriding previously set state from on() / off()')
            self._on = None

        # round to next simulation time step
        dt = nest.GetKernelStatus('resolution')
        value = round(value / dt) * dt

        nest.SetStatus(self._generator, {'start': value})
        self._start = value

    @property
    def stop(self):
        return self._stop

    @rate.setter
    def stop(self, value):
        if self._on is not None:
            warnings.warn('stop: overriding previously set state from on() / off()')
            self._on = None

        # round to next simulation time step
        dt = nest.GetKernelStatus('resolution')
        value = round(value / dt) * dt

        nest.SetStatus(self._generator, {'stop': value})
        self._stop = value

    def on(self):
        if self._start is not None or self._stop is not None:
            warnings.warn('on(): overriding previously set start/stop values')
            self._start = None
            self._stop = None

        nest.SetStatus(self._generator, {'start': 0., 'stop': 1e308})
        self._on = True

    def off(self):
        if self._start is not None or self._stop is not None:
            warnings.warn('off(): overriding previously set start/stop values')
            self._start = None
            self._stop = None

        nest.SetStatus(self._generator, {'start': 0., 'stop': 0.})
        self._on = False

    def is_on(self):
        if self._on is not None:
            return self._on

        return self._start <= nest.GetKernelStatus('time') < self._stop

    @property
    def num_units(self):
        return self._num_units
